#include <math.h>
#include <stdio.h>
#include "transforms_lib.h"

/* ESQ  = eccentricity squared
   DA   = major axis of Earth in meters
   MFAC = offset from ellipsoid */

double ESQ72				= 0.006694317778;
double ESQ84				= 0.006694379990141317;
double DA72					= 6378135.0;
double DA84					= 6378137.0;
double EQUATORIAL_RADIUS_84 = 6378137.0;
double POLAR_RADIUS_84      = 6356752.0; 
double MFAC72				= -28.764;
double MFAC84				= -24.764;
double PI					= 3.14159265359;
double DEGtoRAD				= 0.01745329252;
double RADtoDEG				= 57.295779512896;

/* return the square of the input value */
double square(double value)
{
   return(value*value);
}
/* return the cube of the input value */
double cube(double value)
{
   return(value*value*value);
}
/* accept a value in degrees and return the sin of that value */
double sind(double value)
{
   return(sin(value*DEGtoRAD));  /* convert degrees to radians */
}
/* accept a value in degrees and return the cosine of that value */
double cosd(double value)
{
   return(cos(value*DEGtoRAD));  /* convert degrees to radians */
}
/* accept a value in degrees and return the tangent of that value */
double tand(double value)
{
   return(tan(value*DEGtoRAD));  /* convert degrees to radians */
}
/* accept a value and return the arc sine of that value in degrees */
double asind(double value)
{
   return(asin(value)*RADtoDEG); /* convert radians to degrees */
}
/* accept a value and return the arc cosine of that value in degrees */
double acosd(double value)
{
   return(acos(value)*RADtoDEG); /* convert radians to degrees */
}
/* accept a value and return the arc tangent of that value in degrees */
double atand(double value)
{
   return(atan(value)*RADtoDEG); /* convert radians to degrees */
}
/* accept two values and return the arc tangent 2 of those values in degrees */
double atan2d( double value1,
	       double value2)
{
   return(atan2(value1,value2)*RADtoDEG); /* convert radians to degrees */
}

/* convert e,f,g WGS 84 to lat,lon,alt WGS 84 */
int EFG84toLLA84(FILE *errfile,
				 double ee,
				 double ff,
				 double gg,
				 double *tlat,
				 double *tlon,
				 double *msl)
{
double rgg,gg2,fgg2,fpgg2;
double azin,elin,rgin;
double tmsl,ssgg,ghold;

double gk1 = -square(ESQ84) / (1.0-ESQ84);
double gk2 = -2.0 * ESQ84;
double gk3 = square(DA84) * square(ESQ84);
double gk4 = -(1.0 - ESQ84);
double gk5 = 2.0 * square(DA84) * ESQ84 * (1.0 - ESQ84);
double gk6 = square(DA84) * square(1.0 - ESQ84);
double gk7 = DA84;
double gk8 = 1.0 - ESQ84;
double gk9 = square(DA84);
int pass;
   /*
   printf("(EFG84toLLA84)--start---------------------------------------\n");
   printf("(EFG84toLLA84)input: e=%15.5f f=%15.5f g=%15.5f\n",ee,ff,gg);
   */
   rgg = square(ee) + square(ff);
   gg2 = gk7 * gg * sqrt(gk8/(gk8 * rgg + square(gg) ));
   for (pass=1;pass<=2;pass++)
   {
      fgg2 = gk1 * square(gg2) * square(gg2) +
	     gk2 * gg * cube(gg2) +
	     ( gk3 + gk4 * square(gg) - rgg) * square(gg2) +
	     gk5 * gg * gg2 +
	     gk6 * square(gg);
      fpgg2 = 4.0 * gk1 * cube(gg2) +
	      3.0 * gk2 * gg * square(gg2) +
	      2.0 * (gk3 + gk4 * square(gg) - rgg) * gg2 +
	      gk5 * gg;
      if ( fpgg2 != 0.0 ) gg2 = gg2 - fgg2/fpgg2;
   };

   ghold = ( gk8 * (gk9 * gk8 - square(gg2) ));
   if ( ghold <= 0.0 )
   {
      fprintf(errfile,"(EFG84toLLA84) error: negative or zero sqrt value\n");
      return(1);
   };

   *tlat=atand(gg2/sqrt(ghold));
   if ( gg == 0.0 ) *tlat = 0.0;

   if ( ff == 0.0 )  /* if f=0, longitude will be 180 or 0 */
   {
      if ( ee < 0.0 ) *tlon = 180.0;   /* 180 deg east */
      if ( ee > 0.0 ) *tlon = 0.0;     /* prime meridian */
      if ( ee == 0.0 ) *tlon = 0.0;    /* somewhere on a line between */
				       /* the north and south poles */
   }
   else /* ff != 0.0 */
   {
      *tlon=atand(ee/ff) + 90.0;
      if ( ff > 0.0 ) *tlon = 180.0 - *tlon; /* f > 0, in eastern hemisphere */
      if ( ff < 0.0 ) *tlon = -*tlon;        /* f < 0, in western hemisphere */
   };
   /* during efg conversion, it is possible to get latitude > 90 or < -90 */
   if (*tlat >= 90.0 ) { *tlon = 0.0; *tlat = 90.0; }
   if (*tlat <= -90.0 ) { *tlon = 0.0; *tlat = -90.0; }

   if ( *tlat == 0.0 )         /* if latitude = 0 */
      tmsl = sqrt(rgg) - DA84; /* altitude = sqrt(e**2+f**2) - major axis */
   else
      tmsl = gg / sind(*tlat)
	     - (gk7 * gk8/sqrt(1.0 + (gk8 - 1.0) * ( square(sind(*tlat)) )));

   *msl =  tmsl - MFAC84;  /* subtract out ellipsoid offset */

   /*
   printf("(EFG84toLLA84)output: lat=%13.9f lon=%14.9f alt=%15.5f\n",
	 *tlat,*tlon,*msl);
   printf("(EFG84toLLA84)--stop----------------------------------------\n\n");
   */
   return(0);
}

/* convert lat, lon, mslalt to e,f,g using WGS84 ellipsoid */
int LLA84toEFG84(FILE *errfile,
				 double lat,
				 double lon,
				 double m_alt,
				 double *e,
				 double *f,
				 double *g)

{
double alat,alon,msl,sinlat,coslat,sinlon,coslon,h;

   /*
   printf("(LLA84toEFG84)--start---------------------------------------\n");
   printf("(LLA84toEFG84)input: lat=%13.9f lon=%14.9f alt=%15.5f\n",
	  lat,lon,m_alt);
   */
   if ( (lat > 90.0) || (lat < -90.0) )
   {
      fprintf(errfile,"(LLA84toEFG84) problem: lat>90 \n");
      return(1);
   };

   alat = lat;
   alon = lon;
   msl  = m_alt;      
   sinlat = sind(alat);
   coslat = cosd(alat);
   sinlon = sind(alon);
   coslon = cosd(alon);
   h = msl + MFAC84;

   *e = ((( DA84 * coslat) / (sqrt(1.0 - ESQ84 * square(sinlat) ))) * coslon)
	+ (h * coslat * coslon);
   *f = ((( DA84 * coslat) / (sqrt(1.0 - ESQ84 * square(sinlat) ))) * sinlon)
	+ (h * coslat * sinlon);
   *g = (( DA84 * (1.0 - ESQ84) * sinlat) /
	(sqrt(1.0 - ESQ84 * square(sinlat) ))) + (h * sinlat);
   /*
   printf("(LLA84toEFG84)output: e=%15.5f f=%15.5f g=%15.5f\n",*e,*f,*g);
   printf("(LLA84toEFG84)--stop----------------------------------------\n\n");
   */
   return(0);
}

/* AERtoLLA84 - convert az,el,rg to lat,lon,alt using the
		supplied sitelat,sitelon,sitealt */
int AERtoLLA84( FILE *errfile,
			    double az,
				double el,
				double rg,
				double sitelat,
				double sitelon,
				double sitealt,
				double *aclat,
				double *aclon,
				double *acalt)

{
double the,phi,cthe,sthe,cphi,sphi;
double rgg,gg2,fgg2,fpgg2,ee,ff,gg,xx,yy,zz;
double azin,elin,rgin;
double tmsl,ssgg,mylat,mylon,myalt;
double sitee,sitef,siteg;

   /*
   printf("(AERtoLLA84)--start-----------------------------------------\n");
   printf("(AERtoLLA84)input: az= %15.5f el= %15.5f rg= %15.5f\n",az,el,rg);
   printf("(AERtoLLA84)input: site lat= %13.9f lon= %14.9f alt= %15.5f\n",
	  sitelat,sitelon,sitealt);
   */

   if ( rg == 0.0 )   /* if range in = 0 then nothing can be done */
   {
      *aclat = sitelat;
      *aclon = sitelon;
      *acalt = sitealt;
      fprintf(errfile,"(AERtoLLA84) problem: input range=0 \n");
      return(1);
   };


   the= sitelat;
   phi= sitelon;
   cthe=cosd(the);
   sthe=sind(the);
   cphi=cosd(phi);
   sphi=sind(phi);

   azin= az;
   elin= el;
   rgin= rg;    

   if ( LLA84toEFG84(errfile,sitelat,sitelon,sitealt,&sitee,&sitef,&siteg) !=0 ) return(1);

   /* get relative x,y,z of aircraft */
   xx = rgin * cosd(elin) * sind(azin);
   yy = rgin * cosd(elin) * cosd(azin);
   zz = rgin * sind(elin);

   /* get e,f,g from relative x,y,z */
   ee = -sphi * xx - cphi * sthe * yy + cphi * cthe * zz + sitee;
   ff = cphi * xx - sphi * sthe * yy + sphi * cthe * zz + sitef;
   gg = cthe * yy + sthe * zz + siteg;

   if ( EFG84toLLA84(errfile,ee,ff,gg,&mylat,&mylon,&myalt) !=0 ) return(1);
   *aclat=mylat;
   *aclon=mylon;
   *acalt=myalt;

   /*
   printf("(AERtoLLA84)output: lat=%13.9f lon=%14.9f alt=%15.5f\n",
	  *aclat,*aclon,*acalt);
   printf("(AERtoLLA84)--stop------------------------------------------\n\n");
   */
   return(0);
}

/* convert EFG (WGS 84) coordinates to az,el,rg of specified site */
int EFG84toAER(	FILE *errfile,
				double epos,
				double fpos,
				double gpos,
				double sitelat,
				double sitelon,
				double sitealt,
				double *azout,
				double *elout,
				double *rgout)

{
double ein,fin,gin,sslat,sslon,clon;
double ee,ff,gg,e,f,g,rx,ry,rz,rg,grd;
double coord[9],ecoord,fcoord,gcoord;

   /*
   printf("(EFG84toAER) --start----------------------------------------\n");
   printf("(EFG84toAER) input e=%15.5f f=%15.5f g=%15.5f\n",epos,fpos,gpos);
   printf("(EFG84toAER) input site lat=%13.9f lon=%14.9g alt=%15.5f\n",
	  sitelat,sitelon,sitealt);
   */
   ein = epos;
   fin = fpos;
   gin = gpos;

   sslat = sitelat;
   sslon = sitelon;
   /* rotation matrix */
   coord[8] = sind(sslat);
   coord[7] = cosd(sslat);
   coord[1] = -sind(sslon);
   coord[4] = cosd(sslon);
   coord[2] = -coord[4] * coord[8];
   coord[3] = coord[4] * coord[7];
   coord[5] = coord[1] * coord[8];
   coord[6] = -coord[1] * coord[7];

   if ( LLA84toEFG84(errfile,sitelat,sitelon,sitealt,&ecoord,&fcoord,&gcoord) != 0 ) return(1);

   ee = ein;
   ff = fin;
   gg = gin;
   e = ee - ecoord;
   f = ff - fcoord;
   g = gg - gcoord;
   rx = e * coord[1] + f * coord[4];
   ry = e * coord[2] + f * coord[5] + g * coord[7];
   rz = e * coord[3] + f * coord[6] + g * coord[8];

   rg = sqrt(square(rx) + square(ry) + square(rz));
   *rgout = rg;

   grd = sqrt(square(rx) + square(ry));
   if ( grd != 0 )
   {
      *azout = atan2d(rx,ry);
      if (*azout < 0) *azout=*azout+360.0;
      *elout = atand(rz/grd);
   }
   else
   {
     *azout=0;
     *elout=0;
   };
   /*
   printf("(EFG84toAER) output az=%15.5f el=%15.5f rg=%15.5f\n",
	  *azout,*elout,*rgout);
   printf("(EFG84toAER)--stop------------------------------------------\n\n");
   */
   return(0);
}

/* convert EFG in WGS 84 to EFG in WGS 72 */
int EFG84toEFG72(FILE *errfile,
				 double e84in,
				 double f84in,
				 double g84in,
				 double *e72out,
				 double *f72out,
				 double *g72out)

{
double dscale = 0.0000002263;
double rot_g = 0.000002686;
double de = 0.0;
double df = 0.0;
double dg = 4.5;
double sina,cosa;

   sina = sin(rot_g);
   cosa = cos(rot_g);

   *e72out = (( e84in - de) * cosa + (f84in - df) * sina) / (1.0 + dscale);
   *f72out = (-(e84in - de) * sina + (f84in - df) * cosa) / (1.0 + dscale);
   *g72out = ( g84in - dg ) / (1.0 + dscale);
   return(0);
}

void return_dms(int		lat_lon,
				double	value,
				int		*deg,
				int		*min,
				double	*sec,
				char	*label)
{
int int_deg,int_min,setneg;
double myvalue,real_deg;
  
   if (lat_lon == LAT) *label='N';
   if (lat_lon == LON) *label='E';

   if (value<0.0) {
	   if (lat_lon == LAT) *label='S';
       if (lat_lon == LON) *label='W';
   }

   myvalue=fabs(value);

   int_deg=(int)(myvalue);				/* integer part of degrees */
   *deg=int_deg;

   real_deg=myvalue-(float)(int_deg);	/* real part of degrees */
   int_min = (int)(real_deg*60.0);		/* x60, take the integer part */
   *min = int_min;			

   *sec=(real_deg-((float)(int_min)/60.0))*3600.0;   

}

long chartoint(char *s,
	       int start,
	       int stop)
{
int i,setneg;
long n;
   
   n=0;
   setneg=FALSE;
   if (s[start] == '-') setneg=TRUE;
   for (i=start;i<=stop;i++)
   {
      if ( (s[i] >='0') && (s[i] <= '9') )
	 n=10 * n + (s[i] - '0');
      else
	 n=0;
   }
   if ( setneg ) n = -n;
   return n;
}

void XYZtoAER(FILE *errfile,
			  double xpos,
			  double ypos,
			  double zpos,
			  double *az,
			  double *el,
			  double *rg)
{
double tempflt;

   /* convert x,y,z to azimuth,elevation,range */
   *rg = sqrt( square(xpos) + square(ypos) + square(zpos) );
   if ( *rg != 0.0 )
   {
     tempflt = zpos / *rg;
     *el = asind(tempflt);
   }
   else
   {
     *az = 0.0;
     *el = 0.0;
   };

   if ( ypos == 0.0 )
   {
     if (xpos > 0.0 ) *az = 90.0;
     if (xpos < 0.0 ) *az = 270.0;
     if (xpos == 0.0 ) *az = 0.0;
   }
   else
   {
     tempflt=xpos/ypos;
     *az = atand(tempflt);
     if ( ypos < 0.0 )
     *az = 180.0 + *az;
   };

}

void AERtoXYZ(FILE *errfile,
			  double az,
			  double el,
			  double rg,
			  double *xpos,
			  double *ypos,
			  double *zpos)
{
double groundrg;

   groundrg  = cosd(el) * rg;
   *zpos     = sind(el) * rg;
   *ypos     = cosd(az) * groundrg;
   *xpos     = sind(az) * groundrg;

}

